//
// ubuntu-device-flash - Tool to download and flash devices with an Ubuntu Image
//                       based system
//
// Copyright (c) 2015 Canonical Ltd.
//
// Written by Sergio Schvezov <sergio.schvezov@canonical.com>
//
package main

import (
	"encoding/json"
	"errors"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"os/signal"
	"path/filepath"
	"runtime"
	"syscall"
	"time"

	"github.com/snapcore/snapd/arch"
	"github.com/snapcore/snapd/asserts"
	"github.com/snapcore/snapd/osutil"
	"github.com/snapcore/snapd/progress"
	"github.com/snapcore/snapd/provisioning"
	"github.com/snapcore/snapd/release"
	"github.com/snapcore/snapd/snap"
	"github.com/snapcore/snapd/store"

	"gopkg.in/yaml.v2"
	"launchpad.net/goget-ubuntu-touch/diskimage"
	"launchpad.net/goget-ubuntu-touch/sysutils"
)

type imageFlavor string

const (
	minSizePersonal = 10
	minSizeCore     = 4
)

const (
	rootSizePersonal = 4096
	rootSizeCore     = 1024
)

const (
	flavorPersonal imageFlavor = "personal"
	flavorCore     imageFlavor = "core"
)

func (f imageFlavor) Channel() string {
	return fmt.Sprintf("ubuntu-%s", f)
}

func (f imageFlavor) minSize() int64 {
	switch f {
	case flavorPersonal:
		return minSizePersonal
	case flavorCore:
		return minSizeCore
	default:
		panic("invalid flavor")
	}
}

func (f imageFlavor) rootSize() int {
	switch f {
	case flavorPersonal:
		return rootSizePersonal
	case flavorCore:
		return rootSizeCore
	default:
		panic("invalid flavor")
	}
}

// Snapper holds common options applicable to snappy based images.
type Snapper struct {
	Channel string `long:"channel" description:"Specify the channel to use" default:"stable"`
	Output  string `long:"output" short:"o" description:"Name of the image file to create" required:"true"`

	Development struct {
		Install       []string `long:"install" description:"Install additional packages (can be called multiple times)"`
		DevicePart    string   `long:"device-part" description:"Specify a local device part to override the one from the server"`
		DeveloperMode bool     `long:"developer-mode" description:"Finds the latest public key in your ~/.ssh and sets it up using cloud-init"`
	} `group:"Development"`

	Positional struct {
		Release        string `positional-arg-name:"release" description:"The release to base the image out of (16 or rolling)" required:"true"`
		ModelAssertion string `positional-arg-name:"model-assertion" description:"The model assertion to use" required:"true"`
	} `positional-args:"yes" required:"yes"`

	img             diskimage.CoreImage
	hardware        diskimage.HardwareDescription
	gadget          diskimage.GadgetDescription
	stagingRootPath string

	prepareImagePath string
	modelAssertion   *asserts.Model

	size int64

	flavor imageFlavor
	device string

	customizationFunc []func() error
}

func (s Snapper) sanityCheck() error {
	// we don't want to overwrite the output, people might get angry :-)
	if osutil.FileExists(s.Output) {
		return fmt.Errorf("Giving up, the desired target output file %#v already exists", s.Output)
	}

	if s.size < s.flavor.minSize() {
		return fmt.Errorf("minimum size for %s is %d", s.flavor, s.flavor.minSize())
	}

	if syscall.Getuid() != 0 {
		return errors.New("command requires sudo/pkexec (root)")
	}
	if s.Positional.Release == "15.04" {
		return errors.New("building 15.04 core images is no longer supported. Please use the ppa:snappy-dev/tools 15.04 version of this tool")
	}

	// ensure we error when running on e.g. 14.04 with a sensible
	// error message instead of super strange error later
	if !osutil.FileExists("/bin/systemctl") {
		return errors.New("need '/bin/systemctl to work")
	}

	return nil
}

func snapTargetPathFromSnapFile(src string) (string, error) {
	snapFile, err := snap.Open(src)
	if err != nil {
		return "", err
	}
	var si snap.SideInfo
	// XXX: copied from snapmgr.go
	metafn := src + ".sideinfo"
	if j, err := ioutil.ReadFile(metafn); err == nil {
		if err := json.Unmarshal(j, &si); err != nil {
			return "", fmt.Errorf("cannot read metadata: %s %s\n", metafn, err)
		}
	}
	info, err := snap.ReadInfoFromSnapFile(snapFile, &si)
	if err != nil {
		return "", err
	}

	// local snap
	if info.Revision.Unset() {
		info.Revision = snap.R(-1)
	}

	return info.MountFile(), nil
}

func copyFile(src, dst string) error {
	cmd := exec.Command("cp", "-va", src, dst)
	if o, err := cmd.CombinedOutput(); err != nil {
		return fmt.Errorf("copy failed: %s %s", err, o)
	}
	return nil
}

func (s *Snapper) prepareImage() error {
	var err error
	s.prepareImagePath, err = ioutil.TempDir("", "prepare-image")
	if err != nil {
		return err
	}

	raw, err := ioutil.ReadFile(s.Positional.ModelAssertion)
	if err != nil {
		return err
	}
	as, err := asserts.Decode(raw)
	if err != nil {
		return err
	}
	s.modelAssertion = as.(*asserts.Model)
	arch.SetArchitecture(arch.ArchitectureType(s.modelAssertion.Architecture()))

	os.Setenv("SNAP_REEXEC", "0")
	cmdArgs := []string{
		"snap",
		"prepare-image",
		"--channel", s.Channel,
		s.Positional.ModelAssertion,
		s.prepareImagePath,
	}
	for _, extra := range s.Development.Install {
		cmdArgs = append(cmdArgs, "--extra-snaps="+extra)
	}

	cmd := exec.Command(cmdArgs[0], cmdArgs[1:]...)
	cmd.Stdin = os.Stdin
	cmd.Stdout = os.Stdout
	cmd.Stderr = os.Stderr
	if err := cmd.Run(); err != nil {
		return fmt.Errorf("cannot run snap bootstrap: %s", err)
	}
	return nil
}

func (s *Snapper) install(systemPath string) error {
	tmpSystemPath := filepath.Join(s.prepareImagePath, "image")

	mv := exec.Command("sh", "-c", fmt.Sprintf("cp -rv %s/* %s", tmpSystemPath, systemPath))
	mv.Stdin = os.Stdin
	mv.Stdout = os.Stdout
	mv.Stderr = os.Stderr
	if err := mv.Run(); err != nil {
		return fmt.Errorf("cannot run snap bootstrap: %s", err)
	}

	return nil
}

func (s *Snapper) extractGadget(gadgetPackage string) error {
	fakeGadgetDir := filepath.Join(s.prepareImagePath, "gadget")

	// HORRIBLE - there is always one more circle of hell, never assume
	//            you have reached the end of it yet
	//
	// the new gadget snaps do no have the old gadget specific stuff
	// anymore - however we still need it to create images until we
	// have the new stuff available
	var gadgetMetaYaml string
	switch gadgetPackage {
	case "pc":
		gadgetMetaYaml = compatCanonicalPCamd64
	case "386":
		gadgetMetaYaml = compatCanonicalPCi386
	case "pi2":
		gadgetMetaYaml = compatCanonicalPi2
	case "dragon":
		gadgetMetaYaml = compatCanonicalDragon

	}
	if gadgetMetaYaml != "" {
		if err := ioutil.WriteFile(filepath.Join(fakeGadgetDir, "meta/snap.yaml"), []byte(gadgetMetaYaml), 0644); err != nil {
			return err
		}
	}

	f, err := ioutil.ReadFile(filepath.Join(fakeGadgetDir, "meta/snap.yaml"))
	if err != nil {
		return errors.New("failed to read gadget yaml")
	}
	var gadget diskimage.GadgetDescription
	if err := yaml.Unmarshal([]byte(f), &gadget); err != nil {
		return errors.New("cannot decode gadget yaml")
	}

	s.gadget = gadget
	s.gadget.SetRoot(fakeGadgetDir)
	return nil
}

// Creates a YAML file inside the image that contains metadata relating
// to the installation.
func (s Snapper) writeInstallYaml(bootMountpoint string) error {
	selfPath, err := exec.LookPath(os.Args[0])
	if err != nil {
		return err
	}

	bootDir := ""

	switch s.gadget.Gadget.Hardware.Bootloader {
	// Running systems use a bindmount for /boot/grub, but
	// since the system isn't booted, create the file in the
	// real location.
	case "grub":
		bootDir = "/EFI/ubuntu/grub"
	}

	installYamlFilePath := filepath.Join(bootMountpoint, bootDir, provisioning.InstallYamlFile)

	i := provisioning.InstallYaml{
		InstallMeta: provisioning.InstallMeta{
			Timestamp: time.Now(),
		},
		InstallTool: provisioning.InstallTool{
			Name: filepath.Base(selfPath),
			Path: selfPath,
			// FIXME: we don't know our own version yet :)
			// Version: "???",
		},
		InstallOptions: provisioning.InstallOptions{
			Size:          s.size,
			SizeUnit:      "GB",
			Output:        s.Output,
			Channel:       s.Channel,
			DevicePart:    s.Development.DevicePart,
			DeveloperMode: s.Development.DeveloperMode,
		},
	}

	data, err := yaml.Marshal(&i)
	if err != nil {
		return err
	}

	// the file isn't supposed to be modified, hence r/o.
	return ioutil.WriteFile(installYamlFilePath, data, 0444)
}

func (s *Snapper) bindMount(d string) (string, error) {
	src := filepath.Join(s.img.Writable(), "system-data", d)
	dst := filepath.Join(s.img.System(), d)

	if err := os.MkdirAll(src, 0755); err != nil {
		return "", err
	}
	cmd := exec.Command("mount", "--bind", src, dst)
	if o, err := cmd.CombinedOutput(); err != nil {
		return "", fmt.Errorf("bind mount failed for %s to %s with: %s %v ", src, dst, err, string(o))
	}
	return dst, nil
}

func (s *Snapper) downloadSnap(snapName string) (string, error) {
	if snapName == "" {
		return "", nil
	}
	// if its pointing to a local file, just return that
	if _, err := os.Stat(snapName); err == nil {
		return snapName, nil
	}
	release.Series = s.Positional.Release

	cfg := store.DefaultConfig()
	cfg.StoreID = s.modelAssertion.Store()
	m := store.New(cfg, nil)
	snap, err := m.Snap(snapName, s.Channel, false, snap.R(0), nil)
	if err != nil {
		return "", fmt.Errorf("failed to find os snap: %s", err)
	}
	pb := progress.NewTextProgress()
	tmpName, err := m.Download(snapName, &snap.DownloadInfo, pb, nil)
	if err != nil {
		return "", err
	}
	// rename to the snapid
	baseName := fmt.Sprintf("%s_%s.snap", snap.SnapID, snap.Revision)
	path := filepath.Join(filepath.Dir(tmpName), baseName)
	if err := os.Rename(tmpName, path); err != nil {
		return "", err
	}

	// write out metadata for first boot
	out, err := json.Marshal(snap)
	if err != nil {
		return "", err
	}
	if err := ioutil.WriteFile(path+".sideinfo", []byte(out), 0644); err != nil {
		return "", err
	}

	return path, nil
}

func (s *Snapper) setup() error {
	if s.gadget.PartitionLayout() != "minimal" {
		return fmt.Errorf("only supporting 'minimal' partition layout")
	}

	printOut("Mounting...")
	if err := s.img.Mount(); err != nil {
		return err
	}
	defer func() {
		printOut("Unmounting...")
		if err := s.img.Unmount(); err != nil {
			fmt.Println("WARNING: unexpected issue:", err)
		}
	}()

	printOut("Provisioning...")
	systemPath := s.img.System()

	// setup a fake system
	if err := os.MkdirAll(systemPath, 0755); err != nil {
		return err
	}

	// this is a bit terrible, we need to download the OS
	// mount it, "sync dirs" (see below) and then we
	// will need to download it again to install it properly
	osSnap, err := s.downloadSnap("ubuntu-core")
	if err != nil {
		return err
	}

	// mount os snap
	cmd := exec.Command("mount", osSnap, systemPath)
	if o, err := cmd.CombinedOutput(); err != nil {
		return fmt.Errorf("os snap mount failed with: %s %v ", err, string(o))
	}
	defer exec.Command("umount", systemPath).Run()

	// bind mount all relevant dirs:
	//  - /snap so that we can temporarly mount the kernel
	//  - /var/lib/snapd to put the downloaded snaps in there
	//  - /tmp so that we can create /tmp/root_dev for grub
	for _, d := range []string{"snap", "var/lib/snapd", "tmp"} {
		dst, err := s.bindMount(d)
		if err != nil {
			return err
		}
		defer exec.Command("umount", dst).Run()
	}

	// bind mount /boot/efi
	dst := filepath.Join(systemPath, "/boot/efi")
	cmd = exec.Command("mount", "--bind", s.img.Boot(), dst)
	if o, err := cmd.CombinedOutput(); err != nil {
		return fmt.Errorf("boot bind mount failed with: %s %v ", err, string(o))
	}
	defer exec.Command("umount", dst).Run()
	switch s.gadget.Gadget.Hardware.Bootloader {
	case "grub":
		// grub needs this
		grubUbuntu := filepath.Join(s.img.Boot(), "EFI/ubuntu/grub")
		os.MkdirAll(grubUbuntu, 0755)

		// and /boot/grub
		src := grubUbuntu
		dst = filepath.Join(systemPath, "/boot/grub")
		cmd = exec.Command("mount", "--bind", src, dst)
		if o, err := cmd.CombinedOutput(); err != nil {
			return fmt.Errorf("boot/ubuntu bind mount failed with: %s %v ", err, string(o))
		}
		defer exec.Command("umount", dst).Run()
	case "u-boot":
		src := s.img.Boot()
		dst = filepath.Join(systemPath, "/boot/uboot")
		cmd = exec.Command("mount", "--bind", src, dst)
		if o, err := cmd.CombinedOutput(); err != nil {
			return fmt.Errorf("boot/ubuntu bind mount failed with: %s %v ", err, string(o))
		}
		defer exec.Command("umount", dst).Run()
	}

	if err := s.img.SetupBoot(); err != nil {
		return err
	}

	if err := s.install(systemPath); err != nil {
		return err
	}

	for i := range s.customizationFunc {
		if err := s.customizationFunc[i](); err != nil {
			return err
		}
	}

	return s.writeInstallYaml(s.img.Boot())
}

// deploy orchestrates the priviledged part of the setup
func (s *Snapper) deploy() error {
	// hack to circumvent https://code.google.com/p/go/issues/detail?id=1435
	runtime.GOMAXPROCS(1)
	runtime.LockOSThread()
	if err := sysutils.EscalatePrivs(); err != nil {
		return err
	}
	defer sysutils.DropPrivs()

	printOut("Formatting...")
	if err := s.img.Format(); err != nil {
		return err
	}

	if err := s.setup(); err != nil {
		return err
	}

	return nil
}

func (s Snapper) printSummary() {
	fmt.Println("New image complete")
	fmt.Println("Summary:")
	fmt.Println(" Output:", s.Output)
	fmt.Println(" Architecture:", s.gadget.Architecture())
	fmt.Println(" Channel:", s.Channel)
	fmt.Println(" Version:", globalArgs.Revision)
}

func (s *Snapper) create() (err error) {
	if err := s.sanityCheck(); err != nil {
		return err
	}

	fmt.Println("Calling prepare image")
	if err := s.prepareImage(); err != nil {
		return err
	}

	fmt.Println("Determining gadget configuration")
	if err := s.extractGadget(s.modelAssertion.Gadget()); err != nil {
		return err
	}
	defer os.RemoveAll(s.stagingRootPath)

	// hack to circumvent https://code.google.com/p/go/issues/detail?id=1435
	runtime.GOMAXPROCS(1)
	runtime.LockOSThread()
	if err := sysutils.DropPrivs(); err != nil {
		return err
	}

	switch s.gadget.Gadget.Hardware.Bootloader {
	case "grub":
		legacy := false
		s.img = diskimage.NewCoreGrubImage(s.Output, s.size, s.flavor.rootSize(), s.hardware, s.gadget, legacy, "gpt")
	case "u-boot":
		label := "msdos"
		if s.gadget.Architecture() == archArm64 {
			label = "gpt"
		}
		s.img = diskimage.NewCoreUBootImage(s.Output, s.size, s.flavor.rootSize(), s.hardware, s.gadget, label)
	default:
		return errors.New("no hardware description in Gadget snap")
	}

	printOut("Partitioning...")
	if err := s.img.Partition(); err != nil {
		return err
	}
	defer func() {
		if err != nil {
			os.Remove(s.Output)
		}
	}()

	// Handle SIGINT and SIGTERM.
	go func() {
		ch := make(chan os.Signal)
		signal.Notify(ch, syscall.SIGINT, syscall.SIGTERM)

		for sig := range ch {
			printOut("Received", sig, "... ignoring")
		}
	}()

	// Execute the following code with escalated privs and drop them when done
	if err := s.deploy(); err != nil {
		return err
	}

	if err := s.img.FlashExtra(); err != nil {
		return err
	}

	s.printSummary()

	return nil
}

func isLegacy(release, channel string, revision int) bool {
	if release != "15.04" {
		return false
	}

	switch channel {
	case "edge":
		return revision <= 149
	case "alpha":
		return revision <= 9
	case "stable":
		return revision <= 4
	}

	return false
}
